
/*

    TiMidity -- Experimental MIDI to WAVE converter
    Copyright (C) 1999 Darwin O'Connor

         This program is free software; you can redistribute it and/or modify
         it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
         (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
         GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.

    mcd_c.c: Control driver to allow Timidity to act as an OS/2 Media Control Device DLL
*/
/*#define INCL_MCIOS2*/
#define INCL_OS2MM

#include <stdlib.h>
#include <stdio.h>
#include <stdarg.h>
#define INCL_DOSSEMAPHORES
#define INCL_DOSPROCESS
#define INCL_NOPMAPI
#define INCL_OS2MM
#include <os2.h>
#include <os2me.h>
#include "config.h"
#include "instrum.h"
#include "output.h"
#include "tables.h"
#include "playmidi.h"
#include "readmidi.h"
#include "controls.h"
#include "common.h"
#include "mcd_c.h"

int read_config_file(char *name);
extern char def_instr_name[256];

static void ctl_refresh(void) {}
static void ctl_total_time(int tt);
static void ctl_master_volume(int mv) {}
static void ctl_file_name(char *name);
static void ctl_current_time(int ct) {}
static void ctl_note(int v) {}
static void ctl_program(int ch, int val) {}
static void ctl_volume(int channel, int val) {}
static void ctl_expression(int channel, int val) {}
static void ctl_panning(int channel, int val) {}
static void ctl_sustain(int channel, int val) {}
static void ctl_pitch_bend(int channel, int val) {}
static void ctl_reset(void) {}
static int ctl_open(int using_stdin, int using_stdout);
static void ctl_close(void) {}
static int ctl_read(int32 *valp);
static int cmsg(int type, int verbosity_level, char *fmt, ...);

/**********************************/
/* export the interface functions */

/*#define ctl mcd_control_mode*/

ControlMode mcd_control_mode=
{
  "OS/2 MCD interface", 'm',
  1,0,0,
  ctl_open,dumb_pass_playing_list, ctl_close, ctl_read, cmsg,
  ctl_refresh, ctl_reset, ctl_file_name, ctl_total_time, ctl_current_time,
  ctl_note,
  ctl_master_volume, ctl_program, ctl_volume,
  ctl_expression, ctl_panning, ctl_sustain, ctl_pitch_bend
};

static FILE *infp=stdin, *outfp=stdout; /* infp isn't actually used yet */

typedef struct {
   ULONG param1;
   USHORT userparm;
   HEV finishSEM;
   USHORT message;
   USHORT Notifycode;
   VOID *param2copy;
} tmsginfo;

typedef struct stinstance {
   int32 events, samples;
   MidiEvent *event;
   HMMIO fp;
   USHORT DeviceID;
   ULONG Openparam1;
   HWND Opencallback;
   USHORT OpenUserparm;
   tmsginfo *queue[20];
   int queuesize;
   HMTX queuedoor;
   HEV waitSEM;
   tmsginfo *deleyedmsg;
   TID threadid;
   int timeformat;
   unsigned int maxtime,
      sectime;
   int status;
   USHORT PlayNotifycode;
   int SeekTo;
} tinstance;

tCuePoint CuePoints[50];
int CuePointCount;
tCuePoint PosAdvise;

tinstance myinstance;
FILE *debugfile = NULL;
unsigned int timesize[3] = {0,1000,3000};
MCI_GENERIC_PARMS GenericParms = {0};
unsigned int currtime;
extern MCI_AMP_OPEN_PARMS AmpOpenParms;
HEV pausesem;
int paused;

unsigned int muldiv(unsigned int a, unsigned int b, unsigned int c);

int timetommtime (unsigned int currtime) {
   fprintf(debugfile,"CurrTime: %d, convert %d, sectime: %d, result: %d\n",currtime,timesize[myinstance.timeformat],myinstance.sectime,muldiv(currtime,timesize[myinstance.timeformat],myinstance.sectime));
   return muldiv(currtime,timesize[myinstance.timeformat],myinstance.sectime);
}

int mmtimetotime (unsigned int currtime) {
   return muldiv(currtime,myinstance.sectime,timesize[myinstance.timeformat]);
}

void Waitfor(HEV sem) {
   if (sem) {
      DosWaitEventSem(sem,SEM_INDEFINITE_WAIT);
      DosCloseEventSem(sem);
   }
}

void removemsg(tinstance *pInstance) {
   int loop=1;

   DosRequestMutexSem(pInstance->queuedoor,SEM_INDEFINITE_WAIT);
   while (loop<pInstance->queuesize) {
      pInstance->queue[loop-1]=pInstance->queue[loop];
      loop++;
   }
   pInstance->queuesize--;
   DosReleaseMutexSem(pInstance->queuedoor);
}

int InsertCuePoint(tCuePoint *NewCuePoint) {
   ULONG CuePoint;
   int loop, loop2;

   if (CuePointCount>=50) return MCIERR_CUEPOINT_LIMIT_REACHED;
   CuePoint=NewCuePoint->CuePoint;
   loop=0;
   while (loop<CuePointCount && CuePoints[loop].CuePoint<CuePoint) loop++;
   if (CuePoints[loop].CuePoint<CuePoint) return MCIERR_DUPLICATE_CUEPOINT;
   loop2=CuePointCount-1;
   while (loop2>=loop) {CuePoints[loop2+1]=CuePoints[loop2]; loop2--;}
   CuePoints[loop]=*NewCuePoint;
   return MCIERR_SUCCESS;
}

int DeleteCuePoint(ULONG CueTime) {
   int loop;

   loop=0;
   while (loop<CuePointCount && CuePoints[loop].CuePoint!=CueTime) loop++;
   if (loop<CuePointCount) {
      while (loop<=CuePointCount) {
         CuePoints[loop]=CuePoints[loop+1];
         loop++;
      }
      CuePointCount--;
      return MCIERR_SUCCESS;
   } else return MCIERR_INVALID_CUEPOINT;
}

void SendCue(USHORT MsgType, ULONG CueTime) {
   tCuePoint *CuePoint;
   ULONG rc;

   if (MsgType==MM_MCICUEPOINT) CuePoint=&CuePoints[0];
   else CuePoint=&PosAdvise;

   rc=mdmDriverNotify(myinstance.DeviceID,CuePoint->Callback,MsgType,CuePoint->UserParm,
                      muldiv(CueTime,timesize[MCI_FORMAT_MMTIME],myinstance.sectime));
   fprintf(debugfile,"Notify Return:%d Window: %X CuePoint: %X MsgType %d Callback %X\n PosAdvise: %X",muldiv(CueTime,timesize[MCI_FORMAT_MMTIME],myinstance.sectime),CuePoint->Callback,
           CuePoint,MsgType,PosAdvise.Callback,&PosAdvise);
}

void Finished(tmsginfo *msginfo, HWND handle) {
   if (msginfo->param1 & MCI_WAIT) fprintf(debugfile,"Post finished:%ld\n",DosPostEventSem(msginfo->finishSEM));
   else if (msginfo->param1 & MCI_NOTIFY)
      mdmDriverNotify(myinstance.DeviceID,handle,MM_MCINOTIFY,
                      msginfo->userparm,
                      MAKEULONG(msginfo->message,msginfo->Notifycode));
   if (msginfo->param2copy!=NULL) free(msginfo->param2copy);
   free(msginfo);
}

void *OpenTimidity(tmsginfo *msginfo) {
   ULONG postcount;
   int loop;
   int TranslateChange;
   char timpath[50];
   char *mmbase;
   USHORT rc;
   HMMIO handle;

   myinstance.deleyedmsg=NULL;
   mmbase=getenv("MMBASE");
   if (mmbase) {
      strcpy(timpath,mmbase);
      strcat(timpath,"\\TIMIDITY");
      add_to_pathlist(timpath);
   }
   mmbase=getenv("TIMIDITYDIR");
   if (mmbase) {
      add_to_pathlist(mmbase);
   } 
   read_config_file(CONFIG_FILE);
   play_mode->rate=DEFAULT_RATE;
   init_tables();
   play_mode->open_output();
   ctl->verbosity=2;
   ctl->open(0,0);
   if (!control_ratio) {
      control_ratio = play_mode->rate / CONTROLS_PER_SECOND;
      if(control_ratio<1)
         control_ratio=1;
      else if (control_ratio > MAX_CONTROL_RATIO)
         control_ratio=MAX_CONTROL_RATIO;
   }
   if (*def_instr_name) set_default_instrument(def_instr_name);
   if (myinstance.fp!=NULLHANDLE) {
      myinstance.event=read_midi_file(myinstance.fp,&(myinstance.events),
                                      &(myinstance.samples));
      mmioClose(myinstance.fp,0);
      ctl_total_time(myinstance.samples);
      myinstance.status=MCI_MODE_STOP;
   };
   load_missing_instruments();
   Finished(msginfo,myinstance.Opencallback);
   while (TRUE) {
      DosResetEventSem(myinstance.waitSEM,&postcount); /*Reset count*/
      fprintf(debugfile,"Reset1 %ld\n",postcount);
      if (myinstance.queuesize==0) {
         DosQueryEventSem(myinstance.waitSEM,&postcount); /*Check for activity since check queuesize*/
         fprintf(debugfile,"Reset2 %ld\n",postcount);
         if (postcount==0) {
            fprintf(debugfile,"Waiting Wait\n");
            DosWaitEventSem(myinstance.waitSEM,SEM_INDEFINITE_WAIT);
            fprintf(debugfile,"Finished Wait\n");
         }
      }
      /*      pInstance=(tinstance *)PVOIDFROMMP(qmsg.mp2);*/
      msginfo=myinstance.queue[0];
      fprintf(debugfile,"Message recieved: %d,%d\n",msginfo->message,myinstance.queuesize);
      removemsg(&myinstance);
      switch(msginfo->message) {
      case MCI_PLAY:
         fprintf(debugfile,"Play Now!\n");
         myinstance.status=MCI_MODE_PLAY;
         myinstance.PlayNotifycode=MCI_NOTIFY_SUCCESSFUL;
         play_midi(myinstance.event,myinstance.events,myinstance.samples,muldiv(currtime,play_mode->rate,myinstance.sectime));
         msginfo->Notifycode=myinstance.PlayNotifycode;
fprintf(debugfile,"currtime: %d SeekTo: %d\n",currtime,myinstance.SeekTo);
         if (myinstance.SeekTo!=-1L) {
            currtime=myinstance.SeekTo;
            myinstance.SeekTo=-1L;
         }
         myinstance.status=MCI_MODE_STOP;
         break;
      case MCI_LOAD:
         fprintf(debugfile,"Load Now!\n");
         if (myinstance.status!=MCI_MODE_NOT_READY) {
            free_instruments();
            free(myinstance.event);
         }
         CuePointCount=0;
         PosAdvise.CuePoint=0;
         handle=(HMMIO)((MCI_LOAD_PARMS *)(msginfo->param2copy))->pszElementName;
         myinstance.event=read_midi_file(handle,&(myinstance.events),&(myinstance.samples));
         if (msginfo->param1 & MCI_OPEN_ELEMENT)
            mmioClose(handle,0);
         ctl_total_time(myinstance.samples);
         load_missing_instruments();
         currtime=0;
         myinstance.status=MCI_MODE_STOP;
         break;
      case MCI_CLOSE:
         fprintf(debugfile,"Close Now!\n");
         if (myinstance.status!=MCI_MODE_NOT_READY) {
            free_instruments();
            fprintf(debugfile,"after free intruments\n");
            free(myinstance.event);
            myinstance.status=MCI_MODE_NOT_READY;
         }
         play_mode->close_output();
         fprintf(debugfile,"after play close\n");
         ctl->close();
         fprintf(debugfile,"Before Close SEM\n");
         DosCloseMutexSem(myinstance.queuedoor);
         DosCloseEventSem(myinstance.waitSEM);
         DosCloseEventSem(pausesem);
         fprintf(debugfile,"Past Close SEM\n");
         loop=1; /*Remove any addition queued messages*/
         while (loop<myinstance.queuesize) {
            if (myinstance.queue[loop]->param2copy!=NULL)
               free(myinstance.queue[loop]->param2copy);
            free(myinstance.queue[loop++]);
         }
         fprintf(debugfile,"Past Free msgs\n");
         Finished(msginfo,((PMCI_GENERIC_PARMS)msginfo->param2copy)->hwndCallback);
         fprintf(debugfile,"Ending thread\n");
         fflush(debugfile);
         return NULL;
      }
      Finished(msginfo,((PMCI_GENERIC_PARMS)msginfo->param2copy)->hwndCallback);
      if (myinstance.deleyedmsg!=NULL) {
         Finished(myinstance.deleyedmsg,((PMCI_GENERIC_PARMS)myinstance.deleyedmsg->param2copy)->hwndCallback);
         myinstance.deleyedmsg=NULL;
      }
   }
}

static int ctl_read(int32 *valp) {
   tmsginfo *msginfo;

   while (myinstance.queuesize>0) {
      msginfo=myinstance.queue[0];
      fprintf(debugfile,"Message recieved (running): %d,%d\n",msginfo->message,myinstance.queuesize);
      switch(msginfo->message) {
      case MCI_STOP:
      case MCI_SEEK:
         removemsg(&myinstance);
         myinstance.deleyedmsg=msginfo;
      case MCI_CLOSE:
      case MCI_LOAD:
         myinstance.PlayNotifycode=MCI_NOTIFY_ABORTED;
         return RC_QUIT;
      case MCI_PLAY:
         removemsg(&myinstance);
      }
   }
  return RC_NONE;
}

ULONG mciDriverEntry(tinstance *pInstance, USHORT usMessage, ULONG ulParam1,
                     ULONG ulParam2, USHORT usUserParm) {
   MMDRV_OPEN_PARMS *open_parm;
   MMIOINFO mmioinfo={0,0,NULL,0,16384,NULL,NULL,NULL,NULL,0,0,{0,0,0,0},0,
                             MMIO_NOTRANSLATE,0,NULL,NULLHANDLE};
   tmsginfo *msginfo;
   int param2size;
   ULONG SeekTo;
   tCuePoint NewCuePoint;

   if (!debugfile) {
      debugfile=fopen("c:/tim.log","w");
      setvbuf(debugfile,NULL,_IOLBF,0);
   }
   fprintf(debugfile,"Message Sent:%d\n",usMessage);
   switch (usMessage) {
   case MCI_OPEN:
      fprintf(debugfile,"Open File\n");
      open_parm=(MMDRV_OPEN_PARMS *)ulParam2;
      /*      pInstance=malloc(sizeof(tinstance));*/
      open_parm->pInstance=&myinstance;
      open_parm->usResourceUnitsRequired=1;
      myinstance.Openparam1=ulParam1;
      myinstance.DeviceID=open_parm->usDeviceID;
      myinstance.Opencallback=open_parm->hwndCallback;
      myinstance.OpenUserparm=usUserParm;
      myinstance.fp=NULLHANDLE;
      if (ulParam1 & MCI_OPEN_MMIO) {
         myinstance.fp=(HMMIO)open_parm->pszElementName;
         mmioGetData(myinstance.fp,&mmioinfo,0);
         if (mmioinfo.ulTranslate & MMIO_TRANSLATEDATA) return MCIERR_UNSUPP_FORMAT_MODE;
      } else {
         if (ulParam1 & MCI_OPEN_ELEMENT) {
fprintf(debugfile,"Enter place A: %s\n",open_parm->pszElementName);
            mmioinfo.fccIOProc=mmioStringToFOURCC("MIDI",0);
            myinstance.fp=mmioOpen(open_parm->pszElementName,&mmioinfo,
                                   MMIO_READ | MMIO_DENYWRITE | MMIO_ALLOCBUF);
            if (!myinstance.fp)
               return MCIERR_FILE_NOT_FOUND;
         }
      }
      myinstance.queuesize=0;
      myinstance.timeformat=MCI_FORMAT_MMTIME;
      myinstance.status=MCI_MODE_NOT_READY;
      myinstance.SeekTo=-1L;
      DosCreateMutexSem(NULL,&(myinstance.queuedoor),0,FALSE);
      DosCreateEventSem(NULL,&(myinstance.waitSEM),0,FALSE);
      DosCreateEventSem(NULL,&pausesem,0,FALSE);
      paused=0;
      msginfo=malloc(sizeof(tmsginfo));
      msginfo->param1=ulParam1;
      msginfo->userparm=usUserParm;
      msginfo->message=usMessage;
      msginfo->param2copy=NULL;
      msginfo->Notifycode=MCI_NOTIFY_SUCCESSFUL;
      if (ulParam1 & MCI_WAIT) DosCreateEventSem(NULL,&(msginfo->finishSEM),0,FALSE);
      else msginfo->finishSEM=NULLHANDLE;
      myinstance.threadid=_beginthread(OpenTimidity,NULL,65536,msginfo);
      Waitfor(msginfo->finishSEM);
      break;
   case MCI_GETDEVCAPS:
      if (ulParam1 & MCI_GETDEVCAPS_MESSAGE) {
         switch (((MCI_GETDEVCAPS_PARMS *)(ulParam2))->usMessage) {
         case MCI_OPEN:
         case MCI_GETDEVCAPS:
         case MCI_PLAY:
         case MCI_CLOSE:
         case MCI_STOP:
         case MCI_LOAD:
         case MCI_STATUS:
         case MCI_SET:
         case MCI_PAUSE:
         case MCI_RESUME:
         case MCI_SEEK:
         case MCI_SET_POSITION_ADVISE:
         case MCI_SET_CUEPOINT:
            ((MCI_GETDEVCAPS_PARMS *)(ulParam2))->ulReturn=MCI_TRUE;
         default:
            ((MCI_GETDEVCAPS_PARMS *)(ulParam2))->ulReturn=MCI_FALSE;
         }
         return (MCI_TRUE_FALSE_RETURN << 16) | MCIERR_SUCCESS;
      }
      if (ulParam1 & MCI_GETDEVCAPS_ITEM) {
         fprintf(debugfile,"Dev Caps: %d\n",(((MCI_GETDEVCAPS_PARMS *)(ulParam2))->ulItem));
         switch (((MCI_GETDEVCAPS_PARMS *)(ulParam2))->ulItem) {
         case MCI_GETDEVCAPS_CAN_PLAY:
         case MCI_GETDEVCAPS_CAN_SETVOLUME:
         case MCI_GETDEVCAPS_HAS_AUDIO:
         case MCI_GETDEVCAPS_CAN_STREAM:
         case MCI_GETDEVCAPS_USES_FILES:
            ((MCI_GETDEVCAPS_PARMS *)(ulParam2))->ulReturn=MCI_TRUE;
            return (MCI_TRUE_FALSE_RETURN << 16) | MCIERR_SUCCESS;
         case MCI_GETDEVCAPS_CAN_EJECT:
         case MCI_GETDEVCAPS_CAN_LOCKEJECT:
         case MCI_GETDEVCAPS_CAN_PROCESS_INTERNAL:
         case MCI_GETDEVCAPS_CAN_RECORD:
         case MCI_GETDEVCAPS_CAN_RECORD_INSERT:
         case MCI_GETDEVCAPS_CAN_SAVE:
         case MCI_GETDEVCAPS_HAS_IMAGE:
         case MCI_GETDEVCAPS_HAS_VIDEO:
            ((MCI_GETDEVCAPS_PARMS *)(ulParam2))->ulReturn=MCI_FALSE;
            return (MCI_TRUE_FALSE_RETURN << 16) | MCIERR_SUCCESS;
         case MCI_GETDEVCAPS_DEVICE_TYPE:
            ((MCI_GETDEVCAPS_PARMS *)(ulParam2))->ulReturn=MCI_DEVTYPE_SEQUENCER;
            return (MCI_DEVICENAME_RETURN << 16) | MCIERR_SUCCESS;
         case MCI_GETDEVCAPS_PREROLL_TYPE:
            ((MCI_GETDEVCAPS_PARMS *)(ulParam2))->ulReturn=MCI_PREROLL_NOTIFIED;
            return (MCI_PREROLL_TYPE_RETURN << 16) | MCIERR_SUCCESS;
         case MCI_GETDEVCAPS_PREROLL_TIME:
            ((MCI_GETDEVCAPS_PARMS *)(ulParam2))->ulReturn=0;
            return (MCI_INTEGER_RETURNED << 16) | MCIERR_SUCCESS;
         default:
            return MCIERR_UNSUPPORTED_FLAG;
         }
      }
      break;
   case MCI_STATUS:
      if (ulParam1!=MCI_WAIT & MCI_STATUS_ITEM) return MCIERR_UNSUPPORTED_FLAG;
fprintf(debugfile,"Status Message: %d\n",((MCI_STATUS_PARMS *)(ulParam2))->ulItem);
      switch (((MCI_STATUS_PARMS *)(ulParam2))->ulItem) {
      case MCI_STATUS_MEDIA_PRESENT:
      case MCI_STATUS_READY:
            ((MCI_STATUS_PARMS *)(ulParam2))->ulReturn=MCI_TRUE;
fprintf(debugfile,"Return true\n");
            return (MCI_TRUE_FALSE_RETURN << 16) | MCIERR_SUCCESS;
         break;
/*      case MCI_STATUS_CAN_PASTE:
      case MCI_STATUS_CAN_REDO:
      case MCI_STATUS_CAN_UNDO:*/
      case MCI_STATUS_VIDEO:
            ((MCI_STATUS_PARMS *)(ulParam2))->ulReturn=MCI_FALSE;
            return (MCI_TRUE_FALSE_RETURN << 16) | MCIERR_SUCCESS;
         break;
      case MCI_STATUS_LENGTH:
            ((MCI_STATUS_PARMS *)(ulParam2))->ulReturn=timetommtime(myinstance.maxtime);
            return (MCI_INTEGER_RETURNED << 16) | MCIERR_SUCCESS;
         break;
      case MCI_STATUS_POSITION:
      case MCI_STATUS_POSITION_IN_TRACK:
            ((MCI_STATUS_PARMS *)(ulParam2))->ulReturn=timetommtime(currtime);
            return (MCI_INTEGER_RETURNED << 16) | MCIERR_SUCCESS;
         break;
      case MCI_STATUS_MODE:
            ((MCI_STATUS_PARMS *)(ulParam2))->ulReturn=myinstance.status;
fprintf(debugfile,"Status: %d\n",myinstance.status);
            return (MCI_MODE_RETURN << 16) | MCIERR_SUCCESS;
         break;
      case MCI_STATUS_TIME_FORMAT:
            ((MCI_STATUS_PARMS *)(ulParam2))->ulReturn=myinstance.timeformat;
            return (MCI_TIME_FORMAT_RETURN << 16) | MCIERR_SUCCESS;
         break;
      case MCI_STATUS_AUDIO:
      case MCI_STATUS_VOLUME:
         return mciSendCommand(AmpOpenParms.usDeviceID,MCI_STATUS,ulParam1,(PVOID)ulParam2,usUserParm);
         break;
      default:
fprintf(debugfile,"Return unsupported\n");
         return MCIERR_UNSUPPORTED_FLAG;
      }
      break;
   case MCI_SET:
      fprintf(debugfile,"Set: %X\n",ulParam1);
      if (ulParam1 & MCI_SET_TIME_FORMAT) {
         fprintf(debugfile,"Set TimeFormat: %d\n",((MCI_SET_PARMS *)(ulParam2))->ulTimeFormat);
         myinstance.timeformat=((MCI_SET_PARMS *)(ulParam2))->ulTimeFormat;
      } else if (ulParam1 & MCI_SET_AUDIO) {
         return mciSendCommand(AmpOpenParms.usDeviceID,MCI_SET,ulParam1,(PVOID)ulParam2,usUserParm);
      }
      break;
   case MCI_PAUSE:
      if (myinstance.status==MCI_MODE_NOT_READY || myinstance.status==MCI_MODE_STOP) 
         return MCIERR_SUCCESS;
      paused=1;
      DosResetEventSem(pausesem,&(SeekTo));
      mciSendCommand(AmpOpenParms.usDeviceID,MCI_PAUSE,MCI_WAIT,&GenericParms,0);
      myinstance.status=MCI_MODE_PAUSE;
      break;
   case MCI_RESUME:
      if (myinstance.status==MCI_MODE_NOT_READY || myinstance.status==MCI_MODE_STOP) 
         return MCIERR_SUCCESS;
      paused=0;
      DosPostEventSem(pausesem);
      mciSendCommand(AmpOpenParms.usDeviceID,MCI_RESUME,MCI_WAIT,&GenericParms,0);
      myinstance.status=MCI_MODE_PLAY;
      break;
   case MCI_SEEK:
      if (myinstance.status==MCI_MODE_NOT_READY || myinstance.status==MCI_MODE_STOP) 
         return MCIERR_SUCCESS;
      if (ulParam1 & MCI_TO) SeekTo=mmtimetotime(((PMCI_SEEK_PARMS)ulParam2)->ulTo);
      else if (ulParam1 & MCI_TO_START) SeekTo=0;
      else if (ulParam1 & MCI_TO_END) SeekTo=myinstance.maxtime;
fprintf(debugfile,"Mode: %d SeekTo: %d\n",myinstance.status,myinstance.SeekTo);
      if (myinstance.status==MCI_MODE_PLAY) {
         myinstance.SeekTo=SeekTo;
      } else {
         currtime=SeekTo;
         break;
      }
   case MCI_PLAY:
   case MCI_CLOSE:
   case MCI_STOP:
   case MCI_LOAD:
      msginfo=malloc(sizeof(tmsginfo));
      msginfo->param1=ulParam1;
      msginfo->userparm=usUserParm;
      msginfo->message=usMessage;
      switch(usMessage) {
      case MCI_PLAY:
         if (myinstance.status==MCI_MODE_NOT_READY) {
            free(msginfo);
            return MCIERR_DEVICE_NOT_READY;
         }
         param2size=sizeof(MCI_PLAY_PARMS);
         break;
      case MCI_LOAD:
         param2size=sizeof(MCI_LOAD_PARMS);
         if (ulParam1 & MCI_OPEN_ELEMENT) {
fprintf(debugfile,"Filename: %s\n",((MCI_LOAD_PARMS *)(ulParam2))->pszElementName);
            mmioinfo.fccIOProc=mmioStringToFOURCC("MIDI",0);
            (HMMIO)((MCI_LOAD_PARMS *)(ulParam2))->pszElementName=mmioOpen(((MCI_LOAD_PARMS *)(ulParam2))->pszElementName,&mmioinfo,MMIO_READ | MMIO_DENYWRITE | MMIO_ALLOCBUF);
            if (!((HMMIO)((MCI_LOAD_PARMS *)(ulParam2))->pszElementName))
               return MCIERR_FILE_NOT_FOUND;
         } else {
            mmioGetData((HMMIO)((MCI_LOAD_PARMS *)(ulParam2))->pszElementName,&mmioinfo,0);
            if (mmioinfo.ulTranslate & MMIO_TRANSLATEDATA) return MCIERR_UNSUPP_FORMAT_MODE;
         }
         break;
      case MCI_STOP:
         if (myinstance.status==MCI_MODE_NOT_READY || myinstance.status==MCI_MODE_STOP) {
            free(msginfo);
            return MCIERR_SUCCESS;
         }
      case MCI_SEEK:
         if (myinstance.status==MCI_MODE_NOT_READY) {
            free(msginfo);
            return MCIERR_FILE_NOT_FOUND;
         }
      case MCI_CLOSE:
         param2size=sizeof(MCI_GENERIC_PARMS);
      }
      msginfo->param2copy=memcpy(malloc(param2size),(void *)ulParam2,param2size);
      if ((ulParam1 & MCI_WAIT) && usMessage!=MCI_CLOSE) /*Don't wait with close*/
         DosCreateEventSem(NULL,&(msginfo->finishSEM),0,0);
      else msginfo->finishSEM=NULLHANDLE;
      fprintf(debugfile,"Queueing:\n");
      DosRequestMutexSem(myinstance.queuedoor,SEM_INDEFINITE_WAIT);
      myinstance.queue[(myinstance.queuesize)++]=msginfo;
      DosReleaseMutexSem(myinstance.queuedoor);
      fprintf(debugfile,"Queue, will post\n");
      DosPostEventSem(myinstance.waitSEM);
      Waitfor(msginfo->finishSEM);
/*      if (usMessage==MCI_CLOSE && ulParam1 & MCI_WAIT)
         DosWaitThread(&myinstance.threadid,DCWW_WAIT); /Extra waiting if close*/
      break;
   case MCI_SET_POSITION_ADVISE:
      if (myinstance.status==MCI_MODE_NOT_READY) {
         return MCIERR_FILE_NOT_FOUND;
      }
      if (ulParam1 & MCI_SET_POSITION_ADVISE_ON) {
         PosAdvise.CuePoint=mmtimetotime(((PMCI_POSITION_PARMS)ulParam2)->ulUnits);
         PosAdvise.Callback=((PMCI_POSITION_PARMS)ulParam2)->hwndCallback;
         PosAdvise.UserParm=((PMCI_POSITION_PARMS)ulParam2)->usUserParm;
fprintf(debugfile,"Callback=%d\n",PosAdvise.Callback);
      } else PosAdvise.CuePoint=0;
      break;
   case MCI_SET_CUEPOINT:
      if (myinstance.status==MCI_MODE_NOT_READY) {
         return MCIERR_FILE_NOT_FOUND;
      }
      if (ulParam1 & MCI_SET_CUEPOINT_ON) {
         NewCuePoint.CuePoint=mmtimetotime(((PMCI_CUEPOINT_PARMS)ulParam2)->ulCuepoint);
         NewCuePoint.Callback=((PMCI_CUEPOINT_PARMS)ulParam2)->hwndCallback;
         NewCuePoint.UserParm=((PMCI_CUEPOINT_PARMS)ulParam2)->usUserParm;
         return InsertCuePoint(&NewCuePoint);
      } else return DeleteCuePoint(mmtimetotime(((PMCI_CUEPOINT_PARMS)ulParam2)->ulCuepoint));
   case MCI_ACQUIREDEVICE:
   case MCI_RELEASEDEVICE:
   case MCIDRV_SAVE:
   case MCIDRV_RESTORE:
      break;
   default:
      return MCIERR_UNRECOGNIZED_COMMAND;
  }
  return MCIERR_SUCCESS;
}

static int ctl_open(int using_stdin, int using_stdout)
{
  if (using_stdin && using_stdout)
    infp=outfp=stderr;
  else if (using_stdout)
    outfp=stderr;
  else if (using_stdin)
    infp=stdout;

  mcd_control_mode.opened=1;
  return 0;
}

static void ctl_total_time(int tt) {
      currtime=0;
      myinstance.maxtime=tt;
      myinstance.sectime=play_mode->rate;
      if (!(play_mode->encoding & PE_MONO))
         {myinstance.sectime*=2; myinstance.maxtime*=2;}
      if (play_mode->encoding & PE_16BIT)
         {myinstance.sectime*=2; myinstance.maxtime*=2;}
}

static int cmsg(int type, int verbosity_level, char *fmt, ...)
{
  va_list ap;
  if ((type==CMSG_TEXT || type==CMSG_INFO || type==CMSG_WARNING) &&
      mcd_control_mode.verbosity<verbosity_level)
    return 0;
  va_start(ap, fmt);
  if (!mcd_control_mode.opened)
    {
      vfprintf(debugfile, fmt, ap);
      fprintf(debugfile, "\n");
    }
  else
    {
      vfprintf(debugfile, fmt, ap);
      fprintf(debugfile, "\n");
    }
  va_end(ap);
  return 0;
}

static void ctl_file_name(char *name)
{
  if (mcd_control_mode.verbosity>=0 || mcd_control_mode.trace_playing)
    fprintf(debugfile, "Playing %s\n", name);
}

int main(int argc, char *argv[]) {
   PVOID instance;
   MMDRV_OPEN_PARMS open_parm;
   MCI_PLAY_PARMS play_parm;
   MCI_GENERIC_PARMS generic_parm;
   char junk[10];

   open_parm.pszElementName="C:/SOUND/PLAY/STARWARS.MID";
   mciDriverEntry(NULL,MCI_OPEN,MCI_WAIT,(ULONG)&open_parm,0);
   instance=open_parm.pInstance;
   mciDriverEntry(instance,MCI_PLAY,0,(ULONG)&play_parm,0);
   printf("Type something:");
   scanf("%s\n",junk);
   mciDriverEntry(instance,MCI_STOP,MCI_WAIT,(ULONG)&generic_parm,0);
   mciDriverEntry(instance,MCI_CLOSE,MCI_WAIT,(ULONG)&generic_parm,0);
}
